#include <Servo.h>
#include <AStar32U4.h>
#include <PololuRPiSlave.h>
#include "PinChangeInterrupt.h"

// revive Astar32U4 using avrdude - does not work for me as avrdude does not find the serial port
// PS C:\Program Files (x86)\Arduino\hardware\tools\avr\bin> .\avrdude -C "C:\Program Files (x86)\Arduino\hardware\tools\avr/etc/avrdude.conf" -c avr109 -p atmega32U4 -P COM3 -e

// C:\Users\vili\AppData\Local\Arduino15\packages\pololu-a-star\hardware\avr\4.0.2\bootloaders\caterina

/* This example program shows how to make the A-Star 32U4 Robot
   Controller into a Raspberry Pi I2C slave.  The RPi and A-Star can
   exchange data bidirectionally, allowing each device to do what it
   does best: high-level programming can be handled in a language such
   as Python on the RPi, while the A-Star takes charge of motor
   control, analog inputs, and other low-level I/O.

   The example and libraries are available for download at:

   https://github.com/pololu/pololu-rpi-slave-arduino-library

   You will need the corresponding Raspberry Pi code, which is
   available in that repository under the pi/ subfolder.  The Pi code
   sets up a simple Python-based web application as a control panel
   for your Raspberry Pi robot.
*/

class EncoderQuad
{
  public:
    /** Initialize the encoder.
     ** \param channelA: initial state of channel A, must be 0 or 1
     ** \param channelB: initial state of channel B, must be 0 or 1 */
    void begin(uint8_t channelAState, uint8_t channelBState) {
      mState = (channelBState << 1) | channelAState;
      mTicks = 0;
    }
    inline void tickA(int8_t pinState);
    inline void tickB(int8_t pinState);
    inline int16_t read() {
      return mTicks;
    }
  private:
    uint8_t mState;           // current state of encoder inputs: bit 0 is channel A, bit 1 is channel B
    volatile int16_t mTicks;
};

//static const int8_t quadEncTbl[] PROGMEM = { 0,-1, 1,0, 1,0,0,-1,-1,0,0, 1,0, 1,-1,0 };  // count encoder ticks in reverse direction
static const int8_t quadEncTbl[] PROGMEM =  { 0, 1, -1, 0, -1, 0, 0, 1, 1, 0, 0, -1, 0, -1, 1, 0 };

inline void EncoderQuad::tickA( int8_t newVal )
{
  uint8_t i = mState << 2;        // move the previous channel A and B state into bits 2 and 3
  mState = (mState & 2) | newVal;     // update the current state of channel A and preserve the state of channel B
  i |= mState;              // new state goes into the bits 0 and 1 to get the 4-bit index
  int8_t dir = (int8_t) pgm_read_byte(&quadEncTbl[i]);
  mTicks += dir;
}

inline void EncoderQuad::tickB( int8_t newVal )
{
  uint8_t i = mState << 2;        // move the previous channel A and B state into bits 2 and 3
  mState = (mState & 1) | (newVal << 1); // update the current state of channel B preserving the state of channel A
  i |= mState;              // new state goes into the bits 0 and 1 to get the 4-bit index
  int8_t dir = (int8_t) pgm_read_byte(&quadEncTbl[i]);
  mTicks += dir;
}

static const int8_t PIN_ENC1_A = 8;
static const int8_t PIN_ENC1_B = 11;
static const int8_t PIN_ENC2_A = 15; // SCK
static const int8_t PIN_ENC2_B = 16; // MOSI

EncoderQuad enc1, enc2;

inline void enc1TickA()  {
  uint8_t trigger = getPinChangeInterruptTrigger(digitalPinToPCINT(PIN_ENC1_A));
  enc1.tickA(trigger == RISING ? 1 : 0);
}

inline void enc1TickB()  {
  uint8_t trigger = getPinChangeInterruptTrigger(digitalPinToPCINT(PIN_ENC1_B));
  enc1.tickB(trigger == RISING ? 1 : 0);
}

inline void enc2TickA()  {
  uint8_t trigger = getPinChangeInterruptTrigger(digitalPinToPCINT(PIN_ENC2_A));
  enc2.tickA(trigger == RISING ? 1 : 0);
}

inline void enc2TickB()  {
  uint8_t trigger = getPinChangeInterruptTrigger(digitalPinToPCINT(PIN_ENC2_B));
  enc2.tickB(trigger == RISING ? 1 : 0);
}

// Custom data structure that we will use for interpreting the buffer.
// We recommend keeping this under 64 bytes total.  If you change the
// data format, make sure to update the corresponding code in
// a_star.py on the Raspberry Pi.

struct Data
{
  bool yellow, green, red;
  bool buttonA, buttonB, buttonC;

  int16_t leftMotor, rightMotor;
  uint16_t batteryMillivolts;
  uint16_t analog[6];

  bool playNotes;
  char notes[14];

  // Encoders are unused in this example.
  int16_t leftEncoder, rightEncoder;
};

PololuRPiSlave<struct Data, 5> slave;
PololuBuzzer buzzer;
AStar32U4Motors motors;
AStar32U4ButtonA buttonA;
AStar32U4ButtonB buttonB;
AStar32U4ButtonC buttonC;

void setup()
{
  // Set up the slave at I2C address 20.
  slave.init(20);

  // Play startup sound.
  buzzer.play("v10>>g16>>>c16");

  pinMode(PIN_ENC1_A, INPUT_PULLUP);
  pinMode(PIN_ENC1_B, INPUT_PULLUP);
  pinMode(PIN_ENC2_A, INPUT_PULLUP);
  pinMode(PIN_ENC2_B, INPUT_PULLUP);
  attachPCINT(digitalPinToPCINT(PIN_ENC1_A), enc1TickA, CHANGE);
  attachPCINT(digitalPinToPCINT(PIN_ENC1_B), enc1TickB, CHANGE);
  attachPCINT(digitalPinToPCINT(PIN_ENC2_A), enc2TickA, CHANGE);
  attachPCINT(digitalPinToPCINT(PIN_ENC2_B), enc2TickB, CHANGE);
  enc1.begin(digitalRead(PIN_ENC1_A), digitalRead(PIN_ENC1_B));
  enc2.begin(digitalRead(PIN_ENC2_A), digitalRead(PIN_ENC2_B));

  Serial.begin(115200);
  Serial.print("A* rpi encoders v1.0");
}

void loop()
{ uint32_t lastTime = millis();
  while (1)
  {
    // Call updateBuffer() before using the buffer, to get the latest
    // data including recent master writes.
    slave.updateBuffer();

    // Write various values into the data structure.
    slave.buffer.buttonA = buttonA.isPressed();
    slave.buffer.buttonB = buttonB.isPressed();
    slave.buffer.buttonC = buttonC.isPressed();

    // Change this to readBatteryMillivoltsLV() for the LV model.
    slave.buffer.batteryMillivolts = readBatteryMillivoltsLV();

    for (uint8_t i = 0; i < 6; i++)
    {
      slave.buffer.analog[i] = analogRead(i);
    }

    // READING the buffer is allowed before or after finalizeWrites().
    ledYellow(slave.buffer.yellow);
    ledGreen(slave.buffer.green);
    ledRed(slave.buffer.red);
    motors.setSpeeds(slave.buffer.leftMotor, slave.buffer.rightMotor);

    // Playing music involves both reading and writing, since we only
    // want to do it once.
    static bool startedPlaying = false;

    if (slave.buffer.playNotes && !startedPlaying)
    {
      buzzer.play(slave.buffer.notes);
      startedPlaying = true;
    }
    else if (startedPlaying && !buzzer.isPlaying())
    {
      slave.buffer.playNotes = false;
      startedPlaying = false;
    }

    uint8_t saveSREG = SREG;
    cli();
    slave.buffer.leftEncoder = enc1.read();
    slave.buffer.rightEncoder = enc2.read();
    SREG = saveSREG;  // restore interrupt state

    // When you are done WRITING, call finalizeWrites() to make modified
    // data available to I2C master.
    slave.finalizeWrites();

    uint32_t now = millis();
    if (now - lastTime > 100)
    { Serial.print(slave.buffer.leftEncoder); Serial.print(", ");
      Serial.println(slave.buffer.rightEncoder);
      lastTime = now;
    }
  }
}
